static char rcsid[] = "@(#)$Id: pager_range.c,v 2.7 2016/03/21 20:26:13 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.7 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "pg_range_imp.h"

DEBUG_VAR(DebugIO,__FILE__,"stateio");

#define PAGER_RANGE_magic    0xEB07

struct pager_range {
    unsigned short              magic;    /* PAGER_RANGE_magic */

    struct pager_range *inherited;

    int refcount;
    int range_flags;
    int quote_level;

    int indent;                     /* Absolute value; Adjusted
					from inherted value if 
					PR_INDENT_ABSOLUTE is not set
				     */

    struct pgrng_bfr_binding * stringbuffer;
};

struct pager_range * malloc_simple_pager_range(inherit,range_flags,
					       quote_level,indent)
     struct pager_range *inherit; 
     int range_flags;
     int quote_level;
     int indent;
{
    struct pager_range *ret = safe_zero_alloc(sizeof (*ret));

    ret->magic = PAGER_RANGE_magic;
    ret->refcount = 1;
    ret->range_flags = range_flags;
    ret->quote_level = quote_level;

    ret->indent      = indent;

    if (ret->range_flags < 0 ||
	ret->quote_level < 0 ||
	(ret->indent < 00 && (range_flags & PR_INDENT_ABSOLUTE)))
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "malloc_simple_pager_range",
	      "Bad options",0);

    /* Not actually needed */
    ret->inherited   = inherit;
    if (ret->inherited) {
	inc_pager_range_refcount(ret->inherited);

	ret->range_flags  |= ret->inherited->range_flags;
	ret->quote_level  += ret->inherited->quote_level;       

	if (!(range_flags & PR_INDENT_ABSOLUTE))
	    ret->indent +=  ret->inherited->indent;
    }
   
    ret->stringbuffer  = NULL;

    return ret;
}


void  free_pager_range(range)
     struct pager_range **range;
{
    if (PAGER_RANGE_magic != (*range)->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,"free_pager_range",
	      "Bad magic number",0);

    (*range)->refcount--;

    if ((*range)->refcount > 0) {  /* Do not free */
	
        (*range) = NULL;           /* Refefence count for this pointer is decrement, */
	                           /* so this must have be reset                     */
        return;
    }

    if ((*range)->inherited)
	free_pager_range(& ((*range)->inherited));	

    if ((*range)->stringbuffer)
	pgrng_bfr_bnd_clear_range(& ((*range)->stringbuffer), *range);

    (*range)->magic = 0;    /* Invalidate */
    free(*range);
    *range = NULL;
}

void inc_pager_range_refcount(range)
     struct pager_range *range;
{
    if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,"inc_pager_range_refcount",
	      "Bad magic number",0);

    range->refcount++;
}

/* Caller must call free_pager_range() */
struct pager_range * give_parent_pager_range(range)
     const struct pager_range *range;
{
   if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "give_parent_pager_range",
	      "Bad magic number",0);

     if (range->inherited)
	 inc_pager_range_refcount(range->inherited);

     return range->inherited;
}

int get_pager_range_flags(range)
     const struct pager_range *range;
{
    if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,"get_pager_range_flags",
	      "Bad magic number",0);
    return range->range_flags;
}

/* Quote level of text/plain; format=flowed */
int get_pager_range_quote_level(range)
     const struct pager_range *range;
{
    if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "get_pager_range_quote_level",
	      "Bad magic number",0);

    return range->quote_level;
}

/* Indent column 0 .. (columns)-1  ... may be out of range */
int get_pager_range_indent(range)
     const struct pager_range *range;
{
    if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "get_pager_range_indent",
	      "Bad magic number",0);

    return range->indent;
}

/* range1 and range2 may be NULL */
enum range_changed get_pager_range_change(range1,range2) 
     const struct pager_range * range1;
     const struct pager_range * range2;
{
    enum range_changed ret = rc_same_range;

    if (range1 &&
	PAGER_RANGE_magic != range1->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "get_pager_range_change",
	      "Bad magic number (range1)",0);

    if (range2 &&
	PAGER_RANGE_magic != range2->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "get_pager_range_change",
	      "Bad magic number (range2)",0);

    if (range1 == range2)
	return ret;        /* same range, no changes */
    
    if (range1 && range2) {

	/* No wrapping if direct inheritance and 
	   child have PR_SAME_LINE */

	if (range1 == range2->inherited &&
	    (range2->range_flags & PR_SAME_LINE)) {
	    ret = rc_same_line;
	    goto skip;
	}

	if (range2 == range1->inherited &&
	    (range1->range_flags & PR_SAME_LINE)) {
	    ret = rc_same_line;
	    goto skip;
	}
    }


    /* No wrapping if both have PR_SAME_LINE */

    if (range1) {       
	if (!(range1->range_flags & PR_SAME_LINE))
	    ret = rc_newline;
	else
	    ret = rc_same_line;
    }

    if (range2) {
	if (PAGER_RANGE_magic != range2->magic)
	    panic("PAGER RANGE PANIC",__FILE__,__LINE__,
		  "get_pager_range_change",
		  "Bad magic number (range2)",0);

	if (!(range2->range_flags & PR_SAME_LINE))
	    ret = rc_newline;
	else if (rc_same_range == ret)
	    ret = rc_same_line;

    }

 skip:

    /* changing of quote level requires new line */
    if (range1 && range2 && range1->quote_level != range2->quote_level)
	ret = rc_newline;    
    if (range1 && !range2 && 0 != range1->quote_level)
	ret = rc_newline;
    if (!range1 && range2 && 0 != range2->quote_level)
	ret = rc_newline;

    return ret;
}

/* Caller must call free_pgrng_bfr_binding() */
struct pgrng_bfr_binding * give_pager_range_stringbuffer(range)
     const struct pager_range *range;
{
    if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "give_pager_range_stringbuffer",
	      "Bad magic number",0);

    if (range->stringbuffer)
	inc_pgrng_bfr_binding_refcount(range->stringbuffer);

    return range->stringbuffer;
}


/* Calls pgrng_bfr_bnd_set_range(), caller must handle 
 * free_pgrng_bfr_binding()
 */
void set_pager_range_stringbuffer(range,buffer)
     struct pager_range *range;
     struct pgrng_bfr_binding *buffer;
{
   if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "give_pager_range_stringbuffer",
	      "Bad magic number",0);

   if (range->stringbuffer)
       pgrng_bfr_bnd_clear_range(& (range->stringbuffer), range);
    
    range->stringbuffer = buffer;
    pgrng_bfr_bnd_set_range(buffer,range);
}					    


static int estimate_col P_((const struct string *line));
static int estimate_col(line)
     const struct string *line;
{
    const int tabspacing = 8;
    int L = string_len(line);
    int col = 0;
    int i;

    for ( i = 0; i < L; i++) {
	uint16 ch  = give_unicode_from_string(line,i);	

	switch(ch) {
	    case 0x0009:  /* HT */
		col = ((col / tabspacing) +1) * tabspacing;
		break;

	case 0x000C:   /* FF */
	case 0x000D:   /* CR */
	case 0x000B:   /* VT */
	case 0x000A:   /* LF */
	    col = 0;   /* not always correct */
	    break;
	case UNICODE_SOFT_HYPHEN:
	    break;
	    
	default:
	    col++;    /* not always correct */
	    break;
	}
    }
    
    return col;

}



/* If line is not changed, returns original pointer */

struct string * pager_range_change_line(range,line,filter_on_BOL,filter_on_BOR)
     const struct pager_range *range;
     struct string *line;
     int filter_on_BOL;  /* filter is on beginning of line  */
     int filter_on_BOR;   /* filter is on beginning of range */
{
    /* Used by flush_filter() and flush_filter_so_dir() */
    struct string * result_prefix = NULL;

        
    if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pager_range_change_line",
	      "Bad magic number",0);

    if (filter_on_BOL) {
	/* Original flowed quote */
	int quote_level;
	uint16 first_char =  0x0;

	quote_level = range->quote_level;
	
	if (line && string_len(line) > 0)
	    first_char = give_unicode_from_string(line,0);
	
	if (line) {
	    DPRINT(DebugIO,35,(&DebugIO,
			       "pager_range_change_line: (len=%d) line=%S\n",string_len(line),line));
	}
	
	DPRINT(DebugIO,30,(&DebugIO,
			   "pager_range_change_line: range_flags = %d, quote_level = %d, first_char=04x\n",
			   range->range_flags,
			   quote_level,
			   first_char));
	
	if (0 != (range->range_flags & PR_REPLY_QUOTE))  /* Normal reply quote */
	    quote_level++;
	
	if (quote_level && quote_prefix) {
	    int lastX = -1, i;
	    uint16 repchar = 0x0;
	    int POS = 0;
	    
	    int len = string_len(quote_prefix);
	    
	    DPRINT(DebugIO,35,(&DebugIO, 
			       "pager_range_change_line: (len=%d) prefix=%S\n",
			       len,quote_prefix));
	    
	    for (i = 0; i < len; i++) {
		uint16 ch  = give_unicode_from_string(quote_prefix,i);
		
		if (0 != unicode_ch(ch,UOP_noctrl) &&
		    0 == unicode_ch(ch,UOP_space)) {
		    repchar = ch;
		    lastX = i;
		}
	    }
	    
	    if (len > 0)
		first_char = give_unicode_from_string(quote_prefix,0);
	    
	    if (lastX >= 0) {
		DPRINT(DebugIO,30,(&DebugIO,
				   "pager_range_change_line: lastX=%d, repchar=%04x\n",
				   lastX,repchar));
		
		result_prefix = clip_from_string(quote_prefix,&POS,lastX+1);
		
		if (POS <= lastX) {
		    DPRINT(DebugIO,30,(&DebugIO,
				       "pager_range_change_line: failure, pos=%d\n",
				       POS));
		    goto failure;
		}
	    } else {
		DPRINT(DebugIO,30,(&DebugIO,
				   "pager_range_change_line: No printable characters on prefix\n"));
	    }
	    
	    if (quote_level > 1 && repchar) {
		int x;
		
		if (! result_prefix)
		    result_prefix = new_string(system_charset);
		
		for (x = 0; x < quote_level - 1; x++)
		    add_unicode_to_string(result_prefix,1,&repchar);
	    }
	    
	    if (POS < len) {
		struct string * tmp =  clip_from_string(quote_prefix,&POS,
							len-lastX);
		
		append_string(&result_prefix,tmp,0);
		
		free_string(&tmp);
	    }
	    
	failure:;
	}
	
	if (0 != (range->range_flags & PR_QUOTE_L)) {   /* Quotation for buffer */
	    int need_quote = 0x005B /* [ */ == first_char;
	    
	    DPRINT(DebugIO,30,(&DebugIO,
			       "pager_range_change_line: need_quote=%d first_char=%04x\n",
			       need_quote,first_char));
	    
	    if (need_quote) {
		if (! result_prefix)
		    result_prefix = new_string(system_charset);
		add_unicode_to_string(result_prefix,1,&first_char);		
	    }	
	}
	
	if (range->indent > 0) {
	    int needed = range->indent;
	    
	    /* Not really work */
	    if (result_prefix) {
		int col = estimate_col(result_prefix);
		
		needed -= col;
	    }
	    
	    if (needed > 0) {
		if (! result_prefix)
		    result_prefix = new_string(system_charset);
		fill_ascii_to_string(result_prefix,needed,' ');
	    }
	}
	
    } else if (filter_on_BOR) {
	
	if (range->indent > 0) {

	    DPRINT(DebugIO,35,(&DebugIO,
			       "pager_range_change_line: range indent %d on middle of line not supported\n",
			       range->indent));

	    /* XXXX Need know previous content of line */

	}

    } else {
	DPRINT(DebugIO,35,(&DebugIO,
			   "pager_range_change_line: Not filter_on_BOL or filter_or_BOR\n"));
    }
    
    if (result_prefix) {
	if (line) 
	    append_string(&result_prefix,line,0);
	
	/* Caller frees original string when *res changed */
	
	DPRINT(DebugIO,35,(&DebugIO,
			   "pager_range_change_line: (len=%d) returning changed line=%S\n",
			   string_len(result_prefix),result_prefix));
	return result_prefix;	
    }		  


    DPRINT(DebugIO,35,(&DebugIO,
		       "pager_range_change_line: Returning original line\n"));
    return line;
}

struct pager_range *get_pager_range_serialization_from_file(fh,buffer)
     FILE *fh;
     struct stringbuffer * buffer;
{
    int ch = fgetc(fh);
    int inherited_index = 0;
    int have_inherited = 0;
    int range_flags = 0;
    int quote_level = 0;
    int indent      = 0;
    struct pager_range *ret = NULL;
    struct pager_range *inherited = NULL;

    if ('R' != ch) {
	if (EOF != ch) 
	    ungetc(ch,fh);

	DPRINT(DebugIO,30,(&DebugIO,"get_pager_range_serialization_from_file: Not pager range: %c\n",
			   ch));

	return NULL;
    }

    ch = fgetc(fh);
    if ('I' == ch) {
	have_inherited = 1;

	while (ch >= '0' && ch <= '9') {
	    inherited_index = inherited_index * 10 + ch - '0';
	    ch = fgetc(fh);
	}

	if (',' != ch) {
	    panic("PAGER RANGE PANIC",__FILE__,__LINE__,
		  "get_pager_range_serialization_from_file",
		  "Bad data read from file",0);
	}
	ch = fgetc(fh);

    }

    while (ch >= '0' && ch <= '9') {
	range_flags = range_flags * 10 + ch - '0';
	ch = fgetc(fh);
    }

    if (',' != ch) {
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "get_pager_range_serialization_from_file",
	      "Bad data read from file",0);
    }
    ch = fgetc(fh);

    while (ch >= '0' && ch <= '9') {
	quote_level = quote_level * 10 + ch - '0';
	ch = fgetc(fh);
    }

    if ('i' == ch) {
	int sig = 1;

	ch = fgetc(fh);
	/* Relative values can be negative */

	if ('-' == ch) {
	    sig = -1;
	    ch = fgetc(fh);
	}

	while (ch >= '0' && ch <= '9') {
	    indent = indent * 10 + ch - '0';
	    ch = fgetc(fh);
	}

	indent *= sig;
    }

    if ('\n' != ch) {
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "get_pager_range_serialization_from_file",
	      "Bad data read from file",0);
    }

    /* Recursive call AFTER data is readed */
    if (have_inherited)
	inherited = stringbuffer_get_pager_range(buffer,
						 inherited_index);
    
   
    /* indent is relative if PR_INDENT_ABSOLUTE is not set */

    ret = malloc_simple_pager_range(inherited,range_flags,
				    quote_level,indent);

    if (inherited)
	free_pager_range(&inherited);

    return ret;
}

void put_pager_range_serialization_to_file(fh,range,buffer,pos)
     FILE *fh;
     struct pager_range *range;
     struct stringbuffer *buffer;
     long *pos;
{
    int indent;

    if (PAGER_RANGE_magic != range->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "put_pager_range_serialization_to_file",
	      "Bad magic number",0);

    indent = range->indent;

    if (range->inherited) {
	struct stringbuffer *binherited;
	int index;

	if (PAGER_RANGE_magic != range->inherited->magic)
	    panic("PAGER RANGE PANIC",__FILE__,__LINE__,
		  "put_pager_range_serialization_to_file",
		  "Bad (parent) magic number",0);

	if (! range->inherited->stringbuffer)
	    stringbuffer_add_pager_range(buffer,range->inherited);

	binherited = 
	    pgrng_bfr_bnd_get_stringbuffer(range->inherited->stringbuffer,
					   &index);
	
	/* May write to file */
	if (!binherited)
	    stringbuffer_add_pager_range_b(buffer,
					   range->inherited->stringbuffer,
					   range->inherited);

	if (!binherited)
	    pgrng_bfr_bnd_get_stringbuffer(range->inherited->stringbuffer,
					   &index);
	
	if (binherited != buffer)
	    panic("PAGER RANGE PANIC",__FILE__,__LINE__,
		  "put_pager_range_serialization_to_file",
		  "stringbuffer differ on parent/inherited pager_range",0);

	if (index <0)
	    panic("PAGER RANGE PANIC",__FILE__,__LINE__,
		  "put_pager_range_serialization_to_file",
		  "bad index on parent/inherited pager_range",0);


	if (pos)
	    *pos = ftell(fh);
	fputc('R', fh);
	fprintf(fh,"I%d,", index);

	if (!(range->range_flags & PR_INDENT_ABSOLUTE)) {
	    
	    indent -= range->inherited->indent;
	}

    } else {
	if (pos)
	    *pos = ftell(fh);
	fputc('R', fh);
    }

    fprintf(fh,"%d,%d", 
	    range->range_flags,
	    range->quote_level);

    if (indent != 0) {
	/* This stores relative indent if
	 *  PR_INDENT_ABSOLUTE is not set 
	 */

	fprintf(fh,"i%d",
		indent);		
    }

    fputc('\n', fh);
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
